/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#include <signal.h>
#include <iostream.h>
#include <sys/param.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <sys/utsname.h>

#ifdef __GNUG__
#pragma implementation "interconnect_svc.hxx"
#endif

#include "list.hxx"



#include "sockrpc.hxx"
#include "connectdinfo.hxx"
#include "pollset.hxx"

static void dummy() {
  SvcSocket* svc ;
  ListVal<ConnectdInfo> l1 ;
  ListVal<ServDescr> l2 ;
  *svc << l1 ;
  *svc >> l1 ;
  *svc << l2 ;
  *svc >> l2 ;
}

extern u_int YOOC_PORT ;

#if defined(__alpha) && !defined(__GNUG__)
#include "alphadecl.h"
#endif


#include "connectd_svc.hxx"

static char hostname[256] ;


void ConnectionD::verifyServers()
{
  IterListVal<int> iter(proc_l) ;
  while (iter.more()) {
    if (kill(iter.get(), SIGUSR1) == -1) {
      this->purgeServer(iter.get()) ;
      iter.remove() ;
    } else {
      iter.next() ;
    }
  }  
}


void ConnectionD::purgeServer(int pid)
{
  IterListPtr<ServDescr> iter(serv_l) ;
  while (iter.more()) {
    if (iter.get()->pid() == pid) {
      iter.remove() ;
    } else {
      iter.next() ;
    }
  }  
}


void ConnectionD::connect(int pid) {
  this->proc_l += pid ;
}


void ConnectionD::dispatchCreation(char* host) {
  proxy(ConnectionD) cd ;
  cd.importCless(host, YOOC_PORT) ;
  char name[256] ;
  this->host(name) ;
  cd.notifyDaemon(name) ;
}  


void ConnectionD::dispatchKill(char* kill) {
}  


void MasterConnectionD::dispatchCreation(char* host) {

  IterListPtr<proxy(ConnectionD)> iter(this->daemon_l) ;
  for (; iter.more(); iter.next()) {
    iter.get()->createDaemon(host);
  }
}  


void MasterConnectionD::dispatchKill(char* host) {

  IterListPtr<proxy(ConnectionD)> iter(this->daemon_l) ;
  for (; iter.more(); iter.next()) {
    iter.get()->killDaemon(host);
  }
}  


void MasterConnectionD::stopAllDaemon() {

  IterListPtr<proxy(ConnectionD)> iter(this->daemon_l) ;
  for (; iter.more(); iter.next()) {
    iter.get()->stopDaemon();
  }
}  


void ConnectionD::stopDaemon()
{
  exit(0) ;
}


void MasterConnectionD::stopDaemon()
{
  stopAllDaemon() ;
  exit(0) ;
}


void ConnectionD::infoLocal(ConnectdInfo& cinfo)
{
  this->verifyServers() ;
  this->host(hostname) ;
  cinfo.host = strdup(hostname) ;
  IterListPtr<ServDescr> iter(this->serv_l) ;
  for (; iter.more(); iter.next()) {
    cinfo.serv_l.insertVal(*iter.get(), 0) ;
  }
}


void MasterConnectionD::infoGlobal(ListVal<ConnectdInfo>& cinfo_l)
{
  ConnectdInfo* cinfo = new ConnectdInfo;
  this->infoLocal(*cinfo) ;
  cinfo_l.insertRef(cinfo, 0) ; 
  IterListPtr<proxy(ConnectionD)> iter(this->daemon_l) ;
  for (; iter.more(); iter.next()) {
    cinfo = new ConnectdInfo ;
    iter.get()->infoLocal(*cinfo) ;
    cinfo_l.insertRef(cinfo, 0) ;
  }
}


void ConnectionD::beginServer(char* service,char* filt,u_int port,int pid) 
{
  ServDescr* sd = new ServDescr(service, filt, port, pid) ;
  this->serv_l.insert(sd, 0) ;
}


void ConnectionD::stopServer(char* service, char* filtre, u_int port) 
{
  IterListPtr<ServDescr> iter(this->serv_l) ;
  for (; iter.more(); iter.next()) {
    if (iter.get()->match(service, filtre, port)) {
      delete iter.get() ;
      iter.remove() ;
      break ;
    }
  }
}


void ConnectionD::freeServer(char* service, char* filtre, u_int port) 
{
  IterListPtr<ServDescr> iter(this->serv_l) ;
  for (; iter.more(); iter.next()) {
    if (iter.get()->match(service, filtre, port)) {
      iter.get()->setFree() ;
      break ;
    }
  }
}


int ConnectionD::askLocalServer(char* srv, char* filt, int mode,
				char*& host, u_int& port)
{
  IterListPtr<ServDescr> iter(this->serv_l) ;
  while (iter.more()) {
    if (kill(iter.get()->pid(), SIGUSR1) == -1) {
      delete iter.get() ;
      iter.remove() ;
    } else {      
      if ( (iter.get()->match(srv, filt)) && (iter.get()->isFree()) ) {
	if (mode == 1) iter.get()->setUsed() ;
	port = iter.get()->port() ;
	host = this->host(hostname) ;      
	return 1 ;
      }
      iter.next() ;
    }
  }
  host = NULL ;
  return 0 ;
}


int ConnectionD::askServer(char* serv, char* filt, int mode, char*& host,
			   u_int& port)
{
  if (this->askLocalServer(serv, filt, mode, host, port)) {
    return 1 ;
  }

  IterListPtr<proxy(ConnectionD)> iter(this->daemon_l) ;
  for (; iter.more(); iter.next()) {
    if (iter.get()->askLocalServer(serv, filt, mode, host, port)) {
      return 1 ;
    }
  }
  host = NULL ;  
  return 0 ;
}


void ConnectionD::createDaemon(char* host) 
{
  this->dispatchCreation(host) ;

  proxy(ConnectionD)* cd = new proxy(ConnectionD) ;
  this->daemon_l.insert(cd, 0) ;
  cd->importCless(host, YOOC_PORT) ;
}


void ConnectionD::notifyDaemon(char* host) 
{
  proxy(ConnectionD)* cd = new proxy(ConnectionD) ;
  this->daemon_l.insert(cd, 0) ;
  cd->importCless(host, YOOC_PORT) ;
}


void ConnectionD::killDaemon(char* host) 
{
  IterListPtr<proxy(ConnectionD)> iter(this->daemon_l) ;
  for (; iter.more(); iter.next()) {
    if (strcmp(iter.get()->tohost(hostname), host) == 0) {
      delete iter.get() ;
      iter.remove() ;
    }
  }

  this->dispatchKill(host) ;
}


int ServDescr::match(char* s, char* f, u_int p) 
{
  if ( match(s, f) && (p == _port)) {
    return 1 ;
  } else {
    return 0 ;
  }
}


int ServDescr::match(char* s, char* f) 
{
  if ( (strcmp(this->_service, s) == 0) &&
       !((this->_filter == NULL) && (f != NULL)) && 
       ((f == NULL) || (strcmp(this->_filter, f) == 0)) ) {
    return 1 ;
  } else {
    return 0 ;
  }
}

proxy(ConnectionD) cd ;
ConnectionD* svc ;

void TerminateMaster() {
  ((MasterConnectionD*)svc)->stopAllDaemon() ;
  exit(1) ;
}

void Terminate() {
  char hostname[MAXHOSTNAMELEN] ;
  cd.killDaemon(cd.host(hostname)) ;
  exit(1) ;
}

int main(int argc, char * argv[])
{
  YOOC_PORT = 0 ;
  int vers = 0 ;
  
  for (int i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-p") == 0) {
      i++ ;
      YOOC_PORT = atoi(argv[i]) ;
    } else if (strcmp(argv[i], "-v") == 0) {
      vers = 1 ;
    } else {
      cerr << "usage : " << argv[0] << " [-v] [-p <port number>] " << endl ;
    }
  }

  if (YOOC_PORT == 0) {
    char* port = getenv("YOOC_PORT") ;
    if (port == NULL) {
      cerr << "no specified port number for " << argv[0] << endl ;
      cerr << "use -p option or YOOC_PORT environnement variable" << endl ;
      exit(1) ;
    }
    YOOC_PORT = atoi(port) ;    
  }

  char master[256] ;
  char* dummy = getenv("YOOC_MASTER") ;
  if (dummy == NULL) {
    cerr << "no specified master host" << endl ;
    cerr << "set YOOC_MASTER environnement variable" << endl ;
    exit(1) ;
  } else {
    strcpy(master, getenv("YOOC_MASTER")) ;
  }

  if (vers) {
    printf("YOODA version 1.2 <%s>, Copyright (C) 1994 Eric Abecassis
YOODA comes with ABSOLUTELY NO WARRANTY; for details type `COPYING'
in YOODA_HOME directory. This is free software, and you are welcome to 
redistribute it under certain conditions; for details read `COPYING'
or 'COPYING.LIB' in YOODA_HOME directory\n", argv[0]) ;
  }

  setsid() ;

  struct utsname host ;
  uname(&host); 

  if (strcmp(host.nodename, master) == 0) {
    signal(SIGINT,  ((void (*) (int))TerminateMaster) );
    signal(SIGQUIT, ((void (*) (int))TerminateMaster) );
    signal(SIGTERM, ((void (*) (int))TerminateMaster) );
    signal(SIGSEGV, ((void (*) (int))TerminateMaster) );
    svc = new MasterConnectionD ;
    if (svc->export(YOOC_PORT) == -1) {
      perror("check if another <connectd> is running");
      exit(1) ;
    }
  } else {
    signal(SIGKILL, ((void (*) (int))Terminate) );
    signal(SIGINT,  ((void (*) (int))Terminate) );
    signal(SIGQUIT, ((void (*) (int))Terminate) );
    signal(SIGTERM, ((void (*) (int))Terminate) );
    signal(SIGSEGV, ((void (*) (int))Terminate) );
    if (cd.importCless(master, YOOC_PORT) == -1) {
      cerr << "Master connection daemon not found on " << master << endl ;
      exit(1) ;
    }
    svc = new ConnectionD ;
    svc->notifyDaemon(master) ;
    if (svc->export(YOOC_PORT) == -1) {
      perror("check if another <connectd> is running");
      exit(1) ;
    }
    cd.createDaemon(host.nodename) ;
  }

  PollSet    ss;

  cerr << "connectd daemon bound to " << host.nodename << " "
       << svc->myport() << "\n";

  ss += svc ;
  while (1) {
    ss.execLoop() ;
  }
}

